unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, TeEngine, Series, TeeProcs, Chart, ExtCtrls, RXSpin, Placemnt;

const
  MAX_WIDTH :integer = 10;
  MAX_HEIGHT :integer = 10;
  MAX_CHROM :integer = 100; //MAX_WIDTH * MAX_HEIGHT;

  VALID_FLAG = $8000;
  CHROM_MASK = $7FFF;

type
  EBadCross = Exception;

  TChrom = array of integer;
  TChromArray = array of TChrom;
  TDoubleArray = array of double;
  TDWordArray = array of dword;

  TForm1 = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    CloseButton: TButton;
    Button3: TButton;
    Button1: TButton;
    Chart1: TChart;
    Series1: TFastLineSeries;
    Series2: TFastLineSeries;
    StepCount: TRxSpinEdit;
    RefreshCount: TRxSpinEdit;
    Label1: TLabel;
    PopEdit: TRxSpinEdit;
    PCrossValue: TRxSpinEdit;
    PMutValue: TRxSpinEdit;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    CreateButton: TButton;
    Panel4: TPanel;
    FormStorage1: TFormStorage;
    Label5: TLabel;
    Label6: TLabel;
    GLMax: TLabel;
    Best: TLabel;
    Image1: TImage;
    Button2: TButton;
    Label7: TLabel;
    Generation: TLabel;
    WidthEdit: TRxSpinEdit;
    HeightEdit: TRxSpinEdit;
    Label8: TLabel;
    Label9: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure CreateButtonClick(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure Chart2DblClick(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    Population,NewPopulation:TChromArray;
    pfit: TDoubleArray;
    pcnt: TDWordArray;
    m_psize: dword;
    fitnesses: TDoubleArray;
    best_chrom, chrom_global: TChrom;

    al_pcross,al_pmut: double;
    population_size, generation_counter, glmax_gen, m_lastch: LongInt;

    Iteration,TopIteration: Integer;
    SizeOfTChrom: Integer;

    function RandomDouble(value: double):double;

    procedure RandChrom(var chrom: TChrom);
    procedure ComputeFitness;
    procedure PaintBestChrom;
    procedure Roulette(var chrom: TChrom);
    procedure Ocross(var chrom1,chrom2: TChrom);
    procedure Omut(var chrom: TChrom);
    function Fitness(chrom: Tchrom): double;
    procedure ShowChrom(chrom:TChrom);
  public
    procedure UpdateView;
    procedure OneStep;
    procedure InitPopulation(pop: LongInt; pcross: double; pmut: double);
  end;

var
  Form1: TForm1;

function CreateChrom(var x:TChrom):TChrom;

implementation

uses Math;

{$R *.DFM}

function CreateChrom(var x:TChrom):TChrom;
begin
  SetLength(x,MAX_CHROM);
  Result := x;
end;

procedure TForm1.ComputeFitness;
var j,{best,}worst: integer;
    {max,} min, f_best: double;
begin
  f_best := Fitness(best_chrom);
//  best := 0;
  worst := 0;
  fitnesses[0] := Fitness(Population[0]);
//  max := fitnesses[0];
  min := fitnesses[0];
  for j:=1 to population_size-1 do begin
    fitnesses[j] := Fitness(Population[j]);
{    if fitnesses[j]>max then begin
      max := fitnesses[j];
      best := j;
    end;}
    if fitnesses[j]<min then begin
      min := fitnesses[j];
      worst := j;
    end;
  end;
  move(Population[worst][0],best_chrom[0],sizeofTChrom);
  if (f_best <> min) then m_lastch := generation_counter;

  for j:=0 to population_size-1 do fitnesses[j] := 1/(fitnesses[j]-min+1);
  for j:=1 to population_size-1 do fitnesses[j] := fitnesses[j]+fitnesses[j-1];
end;

function TForm1.Fitness(chrom: Tchrom): double;
var i,len,x1,y1,x2,y2:integer;
begin
  len := 0;

  for i:=0 to MAX_CHROM-2 do begin
    x1 := (integer(chrom[i]) and CHROM_MASK) mod MAX_WIDTH;
    y1 := (integer(chrom[i]) and CHROM_MASK) div MAX_WIDTH;
    x2 := (integer(chrom[i+1]) and CHROM_MASK) mod MAX_WIDTH;
    y2 := (integer(chrom[i+1]) and CHROM_MASK) div MAX_WIDTH;
    len := len + (abs(x1-x2) + abs(y1-y2));
  end;

  x1 := (integer(chrom[MAX_CHROM-1]) and CHROM_MASK) mod MAX_WIDTH;
  y1 := (integer(chrom[MAX_CHROM-1]) and CHROM_MASK) div MAX_WIDTH;
  x2 := (integer(chrom[0]) and CHROM_MASK) mod MAX_WIDTH;
  y2 := (integer(chrom[0]) and CHROM_MASK) div MAX_WIDTH;
  len := len + (abs(x1-x2) + abs(y1-y2));
  Result := len;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Population := nil;
  NewPopulation := nil;
  fitnesses := nil;
  pfit := nil;
  pcnt := nil;
  m_psize := 0;
  generation_counter := 0;
  population_size := 0;
{$IFDEF InitRand}
  randomize;
{$ENDIF}
end;

procedure TForm1.InitPopulation(pop: Integer; pcross, pmut: double);
var i:Integer;
begin
  population_size := pop;
  al_pmut := pmut;
  al_pcross := pcross;
  SetLength(Population,pop);
  SizeOfTChrom := sizeof(integer)*MAX_CHROM;
  for i:=0 to pop-1 do SetLength(Population[i],MAX_CHROM);
  SetLength(NewPopulation,pop);
  for i:=0 to pop-1 do SetLength(NewPopulation[i],MAX_CHROM);
  SetLength(fitnesses,pop);
  SetLength(pfit,pop);
  SetLength(pcnt,pop);
  for i := 0 to pop-1 do RandChrom(Population[i]);

  SetLength(best_chrom,MAX_CHROM);
  SetLength(chrom_global,MAX_CHROM);

  move(Population[0][0],best_chrom[0],sizeofTChrom);
  ComputeFitness;
  generation_counter := 0;
  m_lastch := 0;
  glmax_gen := 0;
  move(best_chrom[0],chrom_global[0],sizeofTChrom);
  UpdateView;
end;

procedure TForm1.Ocross(var chrom1, chrom2: TChrom);
var j1,j2,i1,i2,kde: integer;
    ctmp:TChrom;
    tmp:integer;
    run1,run2: boolean;
begin
  ctmp := CreateChrom(ctmp);
  kde := random(MAX_CHROM);
  move(chrom2[0],ctmp[0],sizeofTChrom);
  move(chrom1[kde],chrom2[kde],sizeofTChrom-kde*sizeof(integer));
  move(ctmp[kde],chrom1[kde],sizeofTChrom-kde*sizeof(integer));
  i1:=0;
  i2:=0;
  j1 := MAX_CHROM;
  j2 := j1;
  while (i1<kde) and (i2<kde) do begin
    run1 := true;
    run2 := true;
    while (i1<kde) and (run1) do begin
      j1:=kde;
      while (j1<MAX_CHROM) and (chrom1[j1]<>chrom1[i1]) do inc(j1);
      if (j1<MAX_CHROM) then run1 := false
      else inc(i1);
    end;
    while (i2<kde) and (run2) do begin
      j2:=kde;
      while (j2<MAX_CHROM) and (chrom2[j2]<>chrom2[i2]) do inc(j2);
      if (j2<MAX_CHROM) then run2 := false
      else inc(i2);
    end;
    if (i1<kde) and (i2<kde) then begin
      if (not (run1 or run2)) then begin
        tmp := chrom1[j1];
        chrom1[j1] := chrom2[j2];
        chrom2[j2] := tmp;
      end else if (not run1) or (not run2) then begin
        raise EBadCross.Create('Beda nam!');
      end;
    end;
    inc(i1);
    inc(i2);
  end;
end;

procedure TForm1.Omut(var chrom: TChrom);
var i,rnd: integer;
    tmp: integer;
    mut:double;
begin
  for i:=0 to MAX_CHROM-1 do begin
    mut:=Random;
    if mut<al_pmut then begin
      rnd := random(MAX_CHROM);
      tmp := chrom[i];
      chrom[i] := chrom[rnd];
      chrom[rnd] := tmp;
    end;
  end;
end;

procedure TForm1.OneStep;
var j: integer;
    tmp: TChromArray;
begin
  j:=0;
  while j<population_size/2 do begin
    Roulette(NewPopulation[2*j]);
    Roulette(NewPopulation[2*j+1]);
    if Random<al_pcross then Ocross(NewPopulation[2*j],NewPopulation[2*j+1]);
    Omut(NewPopulation[2*j]);
    Omut(NewPopulation[2*j+1]);
    inc(j);
  end;
  tmp := Population;
  Population := NewPopulation;
  NewPopulation := tmp;
  ComputeFitness;
  if (Fitness(chrom_global) > Fitness(best_chrom)) then begin
    move(best_chrom[0],chrom_global[0],sizeofTChrom);
    glmax_gen := generation_counter;
  end;
  inc(generation_counter);
  UpdateView;
end;

procedure TForm1.RandChrom(var chrom: TChrom);
var i:integer;
    wch:integer;

function posw(wch:integer; chrom: TChrom):boolean;
var j:integer;
begin
  j:=0;
  while ((j<i) and (chrom[j]<>wch))do
    inc(j);
  Result := i=j;
end;

begin
  fillchar(chrom[0],sizeof(chrom),0);
  for i:=0 to MAX_CHROM-1 do begin
    repeat
      wch := random(MAX_CHROM) or VALID_FLAG;
    until posw(wch,chrom);
    chrom[i] := wch;
  end;
end;

procedure TForm1.Roulette(var chrom: TChrom);
var a: double;
    bottom, top, c: integer;
begin
  a:=RandomDouble(fitnesses[population_size-1]);
  bottom:=0;
  top:=population_size-1;

  while (bottom<top) do begin
    c := (top + bottom) div 2;
    if fitnesses[c]>a then top := c
    else if bottom <> c then  bottom := c else bottom:=c+1;
  end;
  move(Population[top][0],chrom[0],sizeofTChrom);
end;

procedure TForm1.UpdateView;
var a,b:double;
    i:integer;

begin
  a:=0;
  for i:=0 to population_size-1 do a:=a+Fitness(Population[i]);
  a:=a/population_size;
  Series1.Add(a);
  b:=Fitness(best_chrom);
  Series2.Add(b);

  PaintBestChrom;
  
  GLMax.Caption := '['+IntToStr(glmax_gen)+']: ' +  FloatToStr(Fitness(chrom_global));
  Best.Caption := FloatToStr(b);
  Generation.Caption := IntToStr(generation_counter);
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  OneStep;
end;

function TForm1.RandomDouble(value: double): double;
begin
  Result := random * value;
end;

procedure TForm1.Button3Click(Sender: TObject);
var refresh:integer;

begin
  Button1.Enabled := False;
  Button3.Enabled := False;
  TopIteration := Trunc(StepCount.Value);
  refresh := Trunc(RefreshCount.Value);
  Iteration:= 1;
  while Iteration<=TopIteration do begin
    OneStep;
    if (Iteration mod refresh) = 0 then Application.ProcessMessages;
    inc(Iteration);
  end;
  Button1.Enabled := True;
  Button3.Enabled := True;
  UpdateView;
end;

procedure TForm1.CreateButtonClick(Sender: TObject);
begin
  MAX_WIDTH := Trunc(WidthEdit.Value);
  MAX_HEIGHT := Trunc(HeightEdit.Value);
  MAX_CHROM := MAX_WIDTH * MAX_HEIGHT;
  InitPopulation(Trunc(PopEdit.Value),PCrossValue.Value,PMutValue.Value);
  Button1.Enabled := True;
  Button3.Enabled := True;
  Series1.Clear;
  Series2.Clear;
end;

procedure TForm1.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.Chart2DblClick(Sender: TObject);
begin
  ShowChrom(best_chrom);
end;

procedure TForm1.ShowChrom(chrom:TChrom);
var s:string;
    i:integer;
begin
  s := '';
  for i:= 0 to MAX_CHROM-2 do begin
    s := s + IntToStr(chrom[i] and CHROM_MASK) + ';'
  end;
  s:=s+IntToStr(chrom[MAX_CHROM-1] and CHROM_MASK);
  ShowMessage(s);
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  Iteration := TopIteration;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  if population_size>0 then PaintBestChrom;
end;

procedure TForm1.PaintBestChrom;
var i:integer;

function Zero(value:integer):integer;
begin
  Result := Max(0,value);
end;

function XPos(chrom:TChrom):integer;
begin
  Result := ((chrom[i] and CHROM_MASK) mod MAX_WIDTH)*30+20;
end;

function YPos(chrom:TChrom):integer;
begin
  Result := ((chrom[i] and CHROM_MASK) div MAX_WIDTH)*30+Zero((Panel4.Height-(MAX_HEIGHT*30)) div 2)+2;
end;

begin
  i:=0;
  Image1.Canvas.Brush.Color := clBtnFace;
  Image1.Canvas.FillRect(Image1.ClientRect);
  Image1.Canvas.MoveTo(XPos(best_chrom),YPos(best_chrom));
  for i:=0 to MAX_CHROM-1 do begin
    Image1.Canvas.LineTo(XPos(best_chrom),YPos(best_chrom));
  end;
  i:=0;
  Image1.Canvas.MoveTo(XPos(chrom_global)+(MAX_WIDTH*30+10),YPos(chrom_global));
  for i:=0 to MAX_CHROM-1 do begin
    Image1.Canvas.LineTo(XPos(chrom_global)+(MAX_WIDTH*30+10),YPos(chrom_global));
  end;
end;

end.
